#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Module: brokers.virtual.utils
Purpose: Utility functions for virtual broker 
Author: Kieran Mackle
'''

from autotrader.brokers.broker_utils import BrokerUtils

class Utils(BrokerUtils):
    def __init__(self):
        return